/*
 * ircsniff.c - sniffs a port on a server for the irc protocoll
 * by psychoid
 */


#include <stdio.h>
#include <fcntl.h>
#include <netdb.h>
#include <signal.h>
#include <linux/sockios.h>
#include <sys/time.h>
#include <bpf.h>
#include <pcap.h>
#include <unistd.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>
#include <errno.h>
#include <string.h>
#include <getopt.h>
#include <setjmp.h>

#define TH_FIN  0x01
#define TH_SYN  0x02
#define TH_RST  0x04
#define TH_PUSH 0x08
#define TH_ACK  0x10
#define TH_URG  0x20

#define DEBUG_NET 0
#define DEBUG_PARSE 0
#define DEBUG_CHECKOS 0

#define ALARM_PCAP 2
#define ALARM_BLOCKING_IO 3

#define DISP_SRV 1
#define DISP_CLN 2

char deviceglobal[100];
u_long myip;
u_long dest;
int iiport;
int ifrom,ito,all=0;

struct watchlistt {
    char *item;
    struct watchlistt *next;
};

struct clientst {
    char *clientname;
    u_long hostfrom;
    u_short portfrom;
    char labuf[8192];
    struct clientst *next;
};

struct watchlistt *watchlist;
struct clientst *clients;

u_long *pmalloc(int size)
{
    u_long *mem;
    mem=(u_long *)malloc(size);
    if(mem==NULL)
    {
	printf("Error: Out of memory\n");
	exit(0x0);
    }
    memset(mem,0x0,size);
    return mem;
}

char
 *
rlookup (u_long ip)
{
  static char hostname[256];
  struct hostent *host;
  struct sockaddr_in addr;

  addr.sin_addr.s_addr = ip;

  host = gethostbyaddr ((char *) &addr.sin_addr, sizeof (addr.sin_addr), AF_INET);

  if (host == NULL)
    snprintf (hostname,256, "%s\0", inet_ntoa (ip));
  else
    snprintf (hostname, sizeof (hostname), "%s\0", host->h_name);

  return hostname;
}

struct clientst *addclient(char *identifier,u_long fdest, u_short fport)
{
    struct clientst *lku,*llku;
    lku=clients;
    while(lku!=NULL)
    {
	if(lku->hostfrom==fdest && lku->portfrom==fport)
	{
	    goto gotcha;
	}
	llku=lku;
	lku=lku->next;
    }
    if(llku->hostfrom==0)
	lku=llku;
    else
    {
	llku->next=(struct clientst *)pmalloc(sizeof(struct clientst));
	lku=llku->next;
    }
    lku->hostfrom=fdest;
    lku->portfrom=fport;
gotcha:
    if(identifier!=NULL)
    {
	if(lku->clientname!=NULL)
	{
	    if(strlen(lku->clientname)==strlen(identifier))
		if(strstr(lku->clientname,identifier)==lku->clientname)
		    return lku;
	}
	if(lku->clientname!=NULL) free(lku->clientname);
	if(strlen(identifier)<=0) return lku;
	lku->clientname=(char *)pmalloc(strlen(identifier));
	strcpy(lku->clientname,identifier);
	printf("\n%s is %s (%d)\n",identifier,rlookup(fdest),ntohs(fport));
    }
    return lku;
}

struct watchlistt *addwatch(char *what)
{
    struct watchlistt *wlu,*lwlu;
    wlu=watchlist;
    while(wlu!=NULL && wlu->item!=NULL)
    {
	lwlu=wlu;
	wlu=wlu->next;
    }
    if(lwlu->item==NULL)
    {
	wlu=lwlu;
	goto nod;
    }
    wlu=(struct watchlistt *)pmalloc(sizeof(struct watchlistt));
    lwlu->next=wlu;
nod:    
    wlu->item=(char *)pmalloc(strlen(what));
    return wlu;
}

int removewatch(char *what)
{
    struct watchlistt *wlu;
    wlu=watchlist;
    while(wlu!=NULL)
    {
	if(wlu->item!=NULL)
	{
	    if(strlen(wlu->item)==strlen(what))
	    {
		if(strstr(what,wlu->item)==what)
		{
		    free(wlu->item);
		    wlu->item=NULL;
		    return 0x1;
		}
	    }
	}
	wlu=wlu->next;
    }
    return 0x0;
}

int checkwatch(char *what)
{
    struct watchlistt *wlu;
    if(all==1) return 1;
    wlu=watchlist;
    while(wlu!=NULL)
    {
	if(wlu->item!=NULL)
	{
	    if(strlen(wlu->item)==strlen(what))
	    {
		if(strstr(what,wlu->item)==what)
		{
		    return 0x1;
		}
	    }
	}
	wlu=wlu->next;
    }
    return 0x0;
}

int initstructs()
{
    FILE *watch;
    char watt[255];
    struct watchlistt *wat,*lwat;
    watchlist=(struct watchlistt *)pmalloc(sizeof(struct watchlistt));
    clients=(struct clientst *)pmalloc(sizeof(struct clientst));
    watchlist->next=NULL;
    clients->next=NULL;
    watch=fopen("watchlist","r");
    if(watch==NULL)
    {
	printf("No watchlist found. Sniffing them all..\n");
	all=1;
	return 0x0;
    }
    wat=watchlist;
    while(fgets(watt,sizeof(watt),watch))
    {
	sscanf(watt,"%s\n",watt);
	while(wat!=NULL)
	{
	    if(wat->item==NULL) goto nogo;
	    lwat=wat;
	    wat=wat->next;
	}
	lwat->next=(struct watchlistt *)pmalloc(sizeof(struct watchlistt));
	wat=lwat->next;
nogo:
	wat->item=(char *)pmalloc(strlen(watt));
	strcpy(wat->item,watt);
    }
}

struct tcphdr
{
    u_int16_t th_sport;         /* source port */
    u_int16_t th_dport;         /* destination port */
    u_int32_t th_seq;             /* sequence number */
    u_int32_t th_ack;             /* acknowledgement number */
    u_int8_t th_x2:4;           /* (unused) */
    u_int8_t th_off:4;          /* data offset */
    u_int8_t th_flags;
    u_int16_t th_win;           /* window */
    u_int16_t th_sum;           /* checksum */
    u_int16_t th_urp;           /* urgent pointer */
};                                                                                                  

struct ifmap
{
    unsigned long mem_start;
    unsigned long mem_end;
    unsigned short base_addr;
    unsigned char irq;
    unsigned char dma;
    unsigned char port;
};

struct ifreq
{
#define IFHWADDRLEN	 6
#define IFNAMSIZ	 16
    union
    {
	char ifrn_name[IFNAMSIZ];
    } ifr_ifrn;
    
    union {
	struct sockaddr ifru_addr;
	struct sockaddr ifru_dstaddr;
	struct sockaddr ifru_broadaddr;
	struct sockaddr ifru_netmask;
	struct sockaddr ifru_hwaddr;
	short  ifru_flags;
	int    ifru_metric;
	int    ifru_mtu;
	struct ifmap ifru_map;
	char ifru_slave[IFNAMSIZ];
	caddr_t ifru_data;
    } ifr_ifru;
};

#define ifr_name 	ifr_ifrn.ifrn_name
#define ifr_hwaddr 	ifr_ifru.ifru_hwaddr
#define ifr_addr	ifr_ifru.ifru_addr
#define ifr_dstaddr	ifr_ifru.ifru_dstaddr
#define ifr_broadaddr	ifr_ifru.ifru_broadaddr
#define ifr_netmask	ifr_ifru.ifru_netmask
#define ifr_flags	ifr_ifru.ifru_flags
#define ifr_metric	ifr_ifru.ifru_metric
#define ifr_mtu		ifr_ifru.ifru_mtu
#define ifr_map		ifr_ifru.ifru_map
#define ifr_slave	ifr_ifru.ifru_slave
#define ifr_data	ifr_ifru.ifru_data	

struct ifconf
{
    int	ifc_len;
    union
    {
	caddr_t ifcu_buf;
	struct ifreq * ifcu_req;    
    
    } ifc_ifcu;
};

#define ifc_buf ifc_ifcu.ifcu_buf
#define ifc_req ifc_ifcu.ifcu_req

pcap_t *pcap_global_descriptor;
static int offset;

int
startpcap (void)
{
  char errbuf[1028];
  int datalink;
  struct pcap_pkthdr useless;
  if (!deviceglobal || !strcmp (deviceglobal, "default")) {
    strcpy(deviceglobal,pcap_lookupdev (errbuf));
   }
  if (!deviceglobal)
    {
      printf ("Error getting device - %s\n", errbuf);
      exit (1);
    }
  pcap_global_descriptor = pcap_open_live (deviceglobal, 10000, 0, 1000, errbuf);
  if (pcap_global_descriptor == NULL)
    {
      printf ("error opening pcap: %s\n", errbuf);
      printf ("Please enter a valid device (e.g. ppp0 or eth0\n");
      exit (1);
    }

  datalink = pcap_datalink (pcap_global_descriptor);
  bzero (&useless, sizeof (struct pcap_pkthdr));
  switch (datalink)
    {
    case DLT_EN10MB:
      offset = 14;
      break;
    case DLT_NULL:
    case DLT_PPP:
      offset = 4;
      break;
    case DLT_SLIP:
      offset = 16;
      break;
    case DLT_RAW:
      offset = 0;
      break;
    case DLT_SLIP_BSDOS:
    case DLT_PPP_BSDOS:
      offset = 24;
      break;
    case DLT_ATM_RFC1483:
      offset = 8;
      break;
    case DLT_IEEE802:
      offset = 22;
      break;
    default:
      fprintf (stderr, "unknown datalink type (%d)", datalink);
      return (0);
    }
  return 1;
}

void
endpcap (void)
{
  pcap_close (pcap_global_descriptor);
}


int recvtcp( char *buf, int sizeofbuffer)
{
    int datalink;
    struct pcap_pkthdr useless;
    
    struct tcppacket
      {
         struct iphdr ip;
	 struct tcphdr tcp;
	 u_char *data;
      
      } *tcppacket;

    tcppacket = (struct tcppacket *) pcap_next (pcap_global_descriptor, &useless);
    (char *) tcppacket += offset;

    if (ntohs(tcppacket->ip.tot_len) > sizeofbuffer -1)
       return 0;
    memcpy(buf,tcppacket,ntohs(tcppacket->ip.tot_len));
    return 1;
}

u_long getlocalip(void)
{
    int f, count, traverse;
    struct sockaddr_in *blob;
    char buf[512];
    struct ifreq *i;
    struct ifconf ic;
    f = socket(AF_INET,SOCK_STREAM,0);
    ic.ifc_len = 512;
    ic.ifc_buf = buf;
    ioctl(f,SIOCGIFCONF, (char *) &ic);
    i = ic.ifc_req;
    count = ic.ifc_len / sizeof(struct ifreq);
    for (traverse = 0; traverse < count; traverse++)
    {
	ioctl(f,SIOCGIFADDR,(char *) &i);
	blob = (struct sockaddr_in *) & i->ifr_ifru.ifru_addr;
	if (!strcmp(i->ifr_name,deviceglobal))
	   return blob->sin_addr.s_addr;
	i++;
    }
    return 0x0;
}

unsigned short
in_cksum (unsigned short *ptr, int nbytes)
{

  register long sum;
  u_short oddbyte;
  register u_short answer;

  sum = 0;
  while (nbytes > 1)
    {
      sum += *ptr++;
      nbytes -= 2;
    }

  if (nbytes == 1)
    {
      oddbyte = 0;
      *((u_char *) & oddbyte) = *(u_char *) ptr;
      sum += oddbyte;
    }

  sum = (sum >> 16) + (sum & 0xffff);
  sum += (sum >> 16);	
  answer = ~sum;
  return (answer);
}

void
send_raw_tcp (u_long srcip, u_long dstip, u_short th_sport,
	      u_short th_dport, u_short th_flags,
	      u_short th_win)
{

  int sockfd, fromlen;
  int off = 1;
  struct pseudo
    {	
      u_long saddr;
      u_long daddr;
      u_char zero;
      u_char protocol;
      u_short length;
    };
  static struct tcphdr tcpret;
  int retval;
  struct sockaddr_in sheep;

  u_char packet[sizeof (struct iphdr) + sizeof (struct pseudo) + sizeof (struct tcphdr)];
  struct iphdr *ip = (struct iphdr *) packet;
  struct pseudo *pseudo = (struct pseudo *) (packet + sizeof (struct iphdr));
  struct tcphdr *tcp = (struct tcphdr *) (packet + sizeof (struct iphdr) + sizeof (struct pseudo));

  bzero (packet, sizeof (packet));
  bzero (&sheep, sizeof (sheep));

  srand (time (NULL));

  pseudo->saddr = srcip;
  pseudo->daddr = dstip;
  pseudo->zero = 0;
  pseudo->protocol = IPPROTO_TCP;
  pseudo->length = htons (sizeof (struct tcphdr));	/* is this right? */
  ip->version = 4;
  ip->ihl = 5;
  ip->id = 666;
  ip->saddr = srcip;

  ip->daddr = dstip;
  ip->protocol = IPPROTO_TCP;

  ip->ttl = 42;
  ip->frag_off = 0;
  ip->tot_len = sizeof (struct iphdr) + sizeof (struct tcphdr);
  tcp->th_sport = htons (th_sport);
  tcp->th_dport = htons (th_dport);
  tcp->th_seq = htonl (rand ());
  tcp->th_ack = htonl (rand ());
  tcp->th_off = 5;
  tcp->th_flags = th_flags;
  tcp->th_win = htons (th_win);
  tcp->th_urp = 0;
  tcp->th_sum = in_cksum ((u_short *) pseudo, sizeof (struct pseudo) + sizeof (struct tcphdr));

  sockfd = socket (AF_INET, SOCK_RAW, IPPROTO_TCP);
  retval = setsockopt (sockfd, IPPROTO_IP, IP_HDRINCL, &off, sizeof (off));
  memcpy (pseudo, tcp, 20);
  sheep.sin_family = AF_INET;
  sheep.sin_addr.s_addr = dstip;
  fromlen = sizeof (sheep);
  sendto (sockfd, packet, sizeof (struct iphdr) + sizeof (struct tcphdr),
	  0, (struct sockaddr *) &sheep, sizeof (sheep));
  close (sockfd);
}

unsigned long int
lookup (char *hostname)
{
  struct hostent *name;
  unsigned long int address;

  if ((address = inet_addr (hostname)) != -1)
    return address;
  if ((name = gethostbyname (hostname)) == NULL)
    return -1;

  memcpy (&address, name->h_addr, name->h_length);
  return address;
}

void pcap_alarm_handler(int bleh)
{
    send_raw_tcp(inet_addr("1.2.3.4"),inet_addr("1.2.3.4"),0,0,0,0);
    send_raw_tcp(inet_addr("127.0.0.1"),inet_addr("127.0.0.1"),0,0,0,0);
}

char culine[16384];

int processpacket(struct clientst *lku, char *packet)
{
    char *pt;
    pt=strchr(packet,'\n');
    if(pt==NULL)
    {
	*culine=0;
	strcpy(lku->labuf,packet);
	return 0x0;
    }
    else
    {
	*pt++=0;
	memset(culine,0x0,8191);
	if(lku->labuf[0]!=0)
	    strcpy(culine,lku->labuf);
	lku->labuf[0]=0;
	strcat(culine,packet);
	strcpy(packet,pt);
	return 0x1;
    }
}

char dis[10000];

int displayline(struct clientst *lku, int type)
{
    /* avoid double displays */
    char unknown[]="(unknown)";
    char *np;
    if(strstr(dis,culine)==dis && strlen(dis)==strlen(culine)) return 0x0;
    strcpy(dis,culine);
    /* display */
    if(type==DISP_SRV)
    {
	printf("%s\n",culine);
    } else {
	np=lku->clientname;
	if(np==NULL) np=unknown;
	printf("%s!%s@%s(%d) %s\n",np,np,rlookup(lku->hostfrom),ntohs(lku->portfrom),culine);
    }
}

/* from server */

int processinbound(struct clientst *lku, char *packet)
{
    char *pt,*pt1,*pt2,*pt3;
    while(processpacket(lku,packet)==1)
    {
	if(*culine==':')
	{
	    pt=strchr(culine,' ');
	    if(pt!=NULL)
	    {
		*pt=0;
		pt2=strchr(culine,'!');
		*pt=' ';
		if(pt2!=NULL) /* if theres a '!', then its a client talking */
		{
		    pt3=culine+1;
		    *pt2=0;
		    if(checkwatch(pt3)==1)
		    {
			*pt2='!';
			displayline(lku,DISP_SRV);
		    }
		    *pt2='!';
		    pt++;
		    pt1=strchr(pt,' ');
		    if(pt2!=NULL)
		    {
			/* pt has now, what the client does and pt1, to where */
			pt1++;
			pt2=strchr(pt1,' ');
			if(pt2!=NULL)
			{
			    *pt2=0;
			    if(checkwatch(pt1)==1)
			    {
				*pt2=' ';
				displayline(lku,DISP_SRV);
			    }    
			}
		    }
		} else { /* if theres no '!', then theres the server talking */
		    /* this always adresses us, so we advance from server to command to nick */
		    pt=strchr(culine,' ');
		    if(pt!=NULL)
		    {
			/* the command */
			pt++;
			pt=strchr(pt,' ');
			if(pt!=NULL)
			{
			    /* us */
			    pt++;
			    pt2=strchr(pt,' ');
			    if(pt2==NULL) return 0x0;
			    *pt2=0;
			    addclient(pt,lku->hostfrom,lku->portfrom);
			    if(checkwatch(pt)==1)
			    {
				*pt2=' ';
				displayline(lku,DISP_SRV);
			    }
			    *pt2=' ';
			}
		    }
		}
	    }
	}
    }
    return 0x0;
}

/* to server */

int processoutbound(struct clientst *lku, char *packet)
{
    /* this is the data sent by clients */
    while(processpacket(lku,packet)==1)
    {
	if(lku->clientname!=NULL)
	{
	    if(checkwatch(lku->clientname)==1)
		displayline(lku,DISP_CLN);
	}	
    }
}

int irclisten()
{
  int timer = 0, count = 0, portcount = 0;
  char readbuf[8300];
  struct clientst *lku;
  int frk=0;
  u_long inip;
  struct tcppacket
    {
      struct iphdr ip;
      struct tcphdr tcp;
      char data[8192];
    }
   *tcppacket;
  while (1)
    {
      bzero (readbuf, sizeof (readbuf));
      if (recvtcp (readbuf, 8191))
	{
	  tcppacket = (struct tcppacket *) readbuf;
	  if (tcppacket->ip.saddr==dest && ntohs(tcppacket->tcp.th_sport)==iiport)
	  {
	      /* from server */
	      lku=addclient(NULL,tcppacket->ip.daddr,tcppacket->tcp.th_dport);
	      processinbound(lku,tcppacket->data);
	  }
	  if (tcppacket->ip.daddr==dest && ntohs(tcppacket->tcp.th_dport)==iiport)
	  {
	      /* to server */
	      lku=addclient(NULL,tcppacket->ip.saddr,tcppacket->tcp.th_sport);
	      processoutbound(lku,tcppacket->data);		      
	  }
	}
    }
end:
}

int main (int argc, char **argv)
{
    int pid;
    struct sigaction alarmaction;
    bzero(&alarmaction,sizeof(struct sigaction));
    fprintf(stdout,"\nircsniff by psychoid/tCl\n\n",argv[0]);
    if (argc < 4) {
       fprintf(stdout,"Usage: %s ircserver port device(=ppp0/eth0)\n",argv[0]);
       fprintf(stdout,"          you need to be on the same network than the server\n");
       fprintf(stdout,"          Create a file named watchlist, where channels and nicks are stored\n");
       exit(0x0);
    }
    snprintf(deviceglobal,sizeof(deviceglobal),"%s",argv[3]);
    myip =getlocalip();
    if (myip == 0) {
       fprintf(stdout,"No IP found for device %s\n",deviceglobal);
    }    
    fprintf(stdout,"ircserver Host : %s\n",argv[1]);
    dest=lookup(argv[1]);
    if (dest==-1) {
	printf("Destination not found..\n");
	exit (0x0);
    }
    iiport=atoi(argv[2]);
    fprintf(stdout,"to  Port  : %d\n",iiport);
    fprintf(stdout,"\nstarting..\n\n");
    initstructs();
    alarmaction.sa_handler = pcap_alarm_handler;
    alarmaction.sa_flags = 0;
    sigaction(SIGALRM,&alarmaction,NULL);
    startpcap ();
    irclisten();
    endpcap ();
    exit (0x0);
}